package dev.fyloz.trial.colorrecipesexplorer.model

import com.fasterxml.jackson.annotation.JsonIgnore
import dev.fyloz.trial.colorrecipesexplorer.model.validation.NullOrNotBlank
import dev.fyloz.trial.colorrecipesexplorer.model.validation.NullOrSize
import java.time.LocalDate
import java.util.*
import javax.persistence.*
import javax.validation.constraints.NotBlank
import javax.validation.constraints.NotNull
import javax.validation.constraints.Size

private const val RECIPE_ID_NULL_MESSAGE = "Un identifiant est requis"
private const val RECIPE_NAME_NULL_MESSAGE = "Un nom est requis"
private const val RECIPE_DESCRIPTION_NULL_MESSAGE = "Une description est requise"
private const val RECIPE_SAMPLE_NULL_MESSAGE = "Un numéro d'échantillon est requis"
private const val RECIPE_SAMPLE_TOO_SMALL_MESSAGE = "Le numéro d'échantillon doit être supérieur ou égal à 0"
private const val RECIPE_COMPANY_NULL_MESSAGE = "Une bannière est requise"

@Entity
data class Recipe(
    @Id
    @GeneratedValue(strategy = GenerationType.SEQUENCE)
    override val id: Long?,

    /** The name of the recipe. It is not unique in the entire system, but is unique in the scope of a [Company]. */
    val name: String,

    val description: String,

    val sample: Int,

    val approbationDate: LocalDate,

    /** A remark given by the creator of the recipe. */
    val remark: String,

    /** A note that can be altered by anybody with view permissions to the recipe. */
    var note: String,

    @ManyToOne
    val company: Company,

    @JsonIgnore
    @OneToMany
    val mixes: List<Mix>,

    @OneToMany
    var steps: List<RecipeStep>
) : Model {
    constructor(
        id: Long,
        name: String,
        company: Company,
        description: String,
        sample: Int,
        approbationDate: LocalDate,
        remark: String,
        note: String
    ) : this(id, name, description, sample, approbationDate, remark, note, company, listOf(), listOf())

    val mixesSortedById: Collection<Mix>
        @JsonIgnore
        get() = mixes.sortedBy { it.id }

    /** The mix types contained in this recipe. */
    val mixTypes: Collection<MixType>
        @JsonIgnore
        get() = mixes.map { it.getMixType() as MixType }

    /** Checks if the recipe contains the given [mixType]. */
    fun containsMixType(mixType: MixType) = mixTypes.contains(mixType)

    override fun equals(other: Any?): Boolean = other is Recipe && other.name == name && other.company == company
    override fun hashCode(): Int = Objects.hash(name, company)
}

open class RecipeSaveDto(
    @field:NotBlank(message = RECIPE_NAME_NULL_MESSAGE)
    val name: String,

    @field:NotBlank(message = RECIPE_DESCRIPTION_NULL_MESSAGE)
    val description: String,

    @field:NotNull(message = RECIPE_SAMPLE_NULL_MESSAGE)
    @field:Size(min = 0, message = RECIPE_SAMPLE_TOO_SMALL_MESSAGE)
    val sample: Int,

    val approbationDate: LocalDate,

    val remark: String,

    @field:NotNull(message = RECIPE_COMPANY_NULL_MESSAGE)
    val company: Company,
    // TODO when frontend while be done -> add mixes ?
    // TODO when frontend while be done -> add steps ?
) : EntityDto<Recipe> {
    override fun toEntity(): Recipe = recipe(
        name = name,
        description = description,
        sample = sample,
        approbationDate = approbationDate,
        remark = remark,
        company = company
    )
}

open class RecipeUpdateDto(
    @field:NotNull(message = RECIPE_ID_NULL_MESSAGE)
    val id: Long = 0L,

    @field:NullOrNotBlank(message = RECIPE_NAME_NULL_MESSAGE)
    val name: String? = "name",

    @field:NullOrNotBlank(message = RECIPE_DESCRIPTION_NULL_MESSAGE)
    val description: String? = "description",

    @field:NullOrSize(min = 0, message = RECIPE_SAMPLE_TOO_SMALL_MESSAGE)
    val sample: Int? = -1,

    val approbationDate: LocalDate? = LocalDate.MIN,

    val remark: String? = "remark",

    val note: String? = "note",

    val company: Company? = company(),

    val steps: List<RecipeStep>? = listOf()
) : EntityDto<Recipe> {
    override fun toEntity(): Recipe = recipe(
        id,
        name = name ?: "name",
        description = description ?: "description",
        sample = sample ?: -1,
        approbationDate = approbationDate ?: LocalDate.MIN,
        remark = remark ?: "remark",
        note = note ?: "note",
        company = company ?: company(),
        steps = steps ?: listOf()
    )
}

// ==== DSL ====
fun recipe(
    id: Long? = null,
    name: String = "name",
    description: String = "description",
    sample: Int = -1,
    approbationDate: LocalDate = LocalDate.MIN,
    remark: String = "remark",
    note: String = "",
    company: Company = company(),
    mixes: List<Mix> = listOf(),
    steps: List<RecipeStep> = listOf(),
    op: Recipe.() -> Unit = {}
) = Recipe(id, name, description, sample, approbationDate, remark, note, company, mixes, steps).apply(op)

fun recipeSaveDto(
    name: String = "name",
    description: String = "description",
    sample: Int = -1,
    approbationDate: LocalDate = LocalDate.MIN,
    remark: String = "remark",
    company: Company = company(),
    op: RecipeSaveDto.() -> Unit = {}
) = RecipeSaveDto(name, description, sample, approbationDate, remark, company).apply(op)

fun recipeUpdateDto(
    id: Long = 0L,
    name: String = "name",
    description: String = "description",
    sample: Int = -1,
    approbationDate: LocalDate = LocalDate.MIN,
    remark: String = "remark",
    note: String = "",
    company: Company = company(),
    steps: List<RecipeStep> = listOf(),
    op: RecipeUpdateDto.() -> Unit = {}
) = RecipeUpdateDto(id, name, description, sample, approbationDate, remark, note, company, steps).apply(op)
